
/***********************************************************************************
 *   Copyright (C) 2009-2010 by Fabrizio Montesi <famontesi@gmail.com>             *
 *   Copyright (C) 2009-2010 by Ivan Lanese <lanese@cs.unibo.it>                   *
 *   Copyright (C) 2013 by Claudio Guidi <guidiclaudio@gmail.com>                  *
 *   Copyright (C) 2013-2016 by Saverio Giallorenzo <saverio.giallorenzo@gmail.com>*
 *                                                                                 *
 *   This program is free software; you can redistribute it and/or modify          *
 *   it under the terms of the GNU Library General Public License as               *
 *   published by the Free Software Foundation; either version 2 of the            *
 *   License, or (at your option) any later version.                               *
 *                                                                                 *
 *   This program is distributed in the hope that it will be useful,               *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of                *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the                 *
 *   GNU General Public License for more details.                                  *
 *                                                                                 *
 *   You should have received a copy of the GNU Library General Public             *
 *   License along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                               *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.                     *
 *                                                                                 *
 *   For details about the authors of this software, see the AUTHORS file.         *
 ***********************************************************************************/

include "file.iol"
include "runtime.iol"
include "string_utils.iol"
include "console.iol"

include "../config/locations.iol"

include "../__activity_manager/public/interfaces/ActivityManagerInterface.iol"
include "../__activity_manager/public/interfaces/ActivityInterface.iol"
include "../__adaptation_manager/public/interfaces/AdaptationManagerInterface.iol"
include "../__state/public/interfaces/StateInterface.iol"
include "../__roleSupporter/public/interfaces/RoleSupporterInterface.iol"
include "./public/interfaces/AdaptationServerInterface.iol"
include "../environment/environmentInterface.iol"

include "./__rule/public/interfaces/RuleInterface.iol"

execution { concurrent }

outputPort Environment{
  Location: Location_Environment
  Protocol: sodep
  Interfaces: EnvironmentInterface
}

outputPort AdaptationManager {
  Location: Location_AdaptationManager
  Protocol: sodep
  Interfaces: AdaptationManagerInterface, StateInterface
}

outputPort Rule {
  Interfaces: RuleInterface
}

outputPort State {
  Protocol: sodep
  Interfaces: StateInterface
}

outputPort ActivityManager {
  Protocol: sodep
  Interfaces: ActivityManagerInterface
}

inputPort AdaptationServerInput {
  Location: Location_AdaptationServer
  Protocol: sodep
  Interfaces: AdaptationServerInterface
}


outputPort RoleSupporter {
  Location: Location_RoleSupporter
  Protocol: sodep
  Interfaces: RoleSupporterInterface
}

outputPort NewRole {
  Protocol: sodep
  Interfaces: AdaptActivityInterface
}

init
{
  // loads all the rules inside the "rules" folder
  rules -> global.rules;
  registerAdaptationServer@AdaptationManager( global.inputPorts.AdaptationServerInput )();
  getServiceDirectory@File()( serviceDirectory );
  getFileSeparator@File()( fileSeparator );
  rulesDirectory = serviceDirectory + fileSeparator + "rules" + fileSeparator;
  f.directory = rulesDirectory;
  f.regex = ".*Rule\\.ol";
  list@File( f )( list );
  s.type = "Jolie";
  for( i = 0, i < #list.result, i++ ) {
    println@Console( "Loading rule " + list.result[i] )();
    s.filepath = rulesDirectory + list.result[i];
    loadEmbeddedService@Runtime( s )( Rule.location );
    getDescription@Rule()( rules[i] );
    rules[i].location = Rule.location;
    rules[i].filename = list.result[i]
  };
  undef( f )
}

define checkRule {
  println@Console( "Checking rule " + rule.filename )();
  Rule.location = rule.location;
  properties -> rule.properties.name;
  for( j = 0, j < #properties, j++ ){
    eval.properties.( properties[ j ] ) = request.properties.( properties[ j ] ).value
  };
  State.location = request.client;
  stateVariables -> rule.stateVariables.name;
  for( j = 0, j < #stateVariables, j++ ){
    get@State( stateVariables[ j ] )( eval.state.( stateVariables[ j ] ))
  };
  envVariables -> rule.envVariables.name;
  for( j = 0, j < #envVariables, j++ ){
    get@Environment( envVariables[ j ] )( eval.environment.( envVariables[ j ] ))
  };
  evaluateConstraint@Rule( eval )( send );

  // check that the rule does not have any newRole 
  // with the same name of a role in the scope
  for ( newRole in rule.newRoles ) {
    _newRoles.( newRole ) = 1; // <-- used below to exclude new roles
    if( is_defined( request.ports.( newRole ) ) ){
      println@Console( "The rule contains the newRole " + newRole + " already present in the scope" )();
      send = 0
    }
  };

  // check that the rule does not have any non-new role
  // not present in the original scope
  ruleDir = rulesDirectory + rule.activityDirectory;
  f.directory = ruleDir;
  f.dirsOnly = true;
  list@File( f )( list );
  for( r = 0, r < #list.result, r++ ) {
    role = list.result[ r ];
    if( !is_defined( _newRoles.( role ) ) ) {
      println@Console( "Checking the presence of original role " + role )();
      if( !is_defined( request.ports.( role ) )){
        send = 0
      }
    }
  };
  // clean variables
  undef( ruleDir ); undef( _newRoles ); undef( f ); undef( list ); undef( role )
}

define sendRule { 
  println@Console( "\tPreparing the new code" )();
  // ruleDir = rulesDirectory + rules[ i ].activityDirectory;
  ruleDir = rulesDirectory + rule.activityDirectory;
  println@Console( "Selected: " + ruleDir )();

  // loads the emptyProcess code of the rule
  readFile.filename = ruleDir + fileSeparator + "emptyProcess.ol";
  readFile@File( readFile )( emptyProcess );

  // assigns a fresh key to the main scope of the rule
  main_key = response.main_key = new;

  // remaps statics keys of subscopes with fresh ones
  subscopes -> rule.subscopes.name;
  for( subIndex = 0, subIndex < #subscopes, subIndex++ ){
    subscopesMap.( subscopes[ subIndex ] ) = new
  };

  // retrieves the roles involved in this rule (one per folder)
  f.directory = ruleDir;
  f.dirsOnly = true;
  list@File( f )( list );
  println@Console( "Directories retrieved: " + #list.result )();
  undef( f );

  // creates the list of projected roles
  for( r = 0, r < #list.result, r++ ){
    pRoles.( list.result[ r ] ) = 1
  };

  // create the newRoles and store their locations
  for ( newRole in rule.newRoles ) {
    if( is_defined( newRole.location ) ) {
      RoleSupporter.location = newRole.location
    };
    roleName = newRole;
    createRole@RoleSupporter( roleName )( roleID );
    newRolesLocations.( newRole ) = RoleSupporter.location + "/!/" + roleID;
    RoleSupporter.location = Location_RoleSupporter
  };
  
  // send the code to the newRoles
  for ( newRole in rule.newRoles ) {

    rci = 0; // role code counter
    // retrieves the "main" adapted scope and the message-handler
    sf.directory = ruleDir + fileSeparator + newRole;
    println@Console( "Entering " + sf.directory )();
    sf.regex = ".*\\.ol";
    list@File( sf )( fileList );
    for( fileIndex = 0, fileIndex < #fileList.result, fileIndex++ ){
      file = fileList.result[ fileIndex ];
      println@Console( "Loading file " + role + fileSeparator + file )();
      readFile.filename = sf.directory + fileSeparator + file;
      readFile@File( readFile )( code );

      response.roles.( newRole ).code[ rci ].key = main_key;    

      if( file == "mh.ol" ){
        response.roles.( newRole ).code[ rci ].mh = code
      } else {
        // if it is the scope, it injects a fresh key related to that scope
        code.replacement = main_key;
        code.regex = "Adapt__KPH__";
        replaceAll@StringUtils( code )( code );
        // replaces the address of the scope leader
        code.replacement = request.client;
        code.regex = "RPH__ADAPTATION_LEADER";
        replaceAll@StringUtils( code )( code );
        
        // replaces the addresses of existing roles
        foreach ( p : request.ports ){
          code.replacement = request.ports.( p ).address;
          code.regex = "RPH__" + p;
          replaceAll@StringUtils( code )( code )
        };
        // replaces the addresses of the newRoles
          foreach ( p : newRolesLocations ){
            code.replacement = newRolesLocations.( p );
            code.regex = "RPH__" + p + "!";
            replaceAll@StringUtils( code )( code )
          };
        // replaces the addresses of the newRoles if scope leaders
          foreach ( p : newRolesLocations ){
            code.replacement = newRolesLocations.( p );
            code.regex = "RPH__" + p;
            replaceAll@StringUtils( code )( code )
          };
          
        // replaces the static UUID keys with the fresh ones
        foreach( staticScopeUUID : subscopesMap ){
          code.replacement = subscopesMap.( staticScopeUUID );
          code.regex = staticScopeUUID;
          replaceAll@StringUtils( code )( code )
        };
        response.roles.( newRole ).code[ rci ].code = code
      }
    };

    // retrieves the subscopes and the message-handler
    undef( sf.regex );
    sf.dirsOnly = true;
    list@File( sf )( subscopesList );
    undef( sf.dirsOnly );
    sf.regex = ".*\\.ol";
    scopeDirectory = sf.directory;
    for( ssIndex = 0, ssIndex < #subscopesList.result, ssIndex++ ){
      ssFolder = subscopesList.result[ ssIndex ];
      println@Console( "Entering subscope " + ssFolder )();
      sf.directory = scopeDirectory + fileSeparator + ssFolder;
      list@File( sf )( fileList );
      rci++;
      for( fileIndex = 0, fileIndex < #fileList.result, fileIndex++ ){
        file = fileList.result[ fileIndex ];
        println@Console( "Loading file " + sf.directory + fileSeparator + file )();
        readFile.filename = sf.directory + fileSeparator + file;
        readFile@File( readFile )( code );

        response.roles.( newRole ).code[ rci ].key = subscopesMap.( ssFolder );

        if( file == "mh.ol" ){
          response.roles.( newRole ).code[ rci ].mh = code
        } else {
          // replaces the static UUID key with the fresh one
          foreach( staticScopeUUID : subscopesMap ){
            println@Console( "replacing key: " + staticScopeUUID + " with " + subscopesMap.( staticScopeUUID ) )();
            code.replacement = subscopesMap.( staticScopeUUID );
            code.regex = staticScopeUUID;
            replaceAll@StringUtils( code )( code )
          };
          // replaces the addresses of the existing roles
          foreach ( p : request.ports ) {
            code.replacement = request.ports.( p ).address;
            code.regex = "RPH__" + p;
            println@Console( "replacing address: " + code.regex + " with " + code.replacement )();
            replaceAll@StringUtils( code )( code )
          };
          // replaces the addresses of the newRoles
          foreach ( p : newRolesLocations ){
            code.replacement = newRolesLocations.( p );
            code.regex = "RPH__" + p + "!";
            replaceAll@StringUtils( code )( code )
          };          
          // replaces the addresses of the newRoles if scope leaders
          foreach ( p : newRolesLocations ){
            code.replacement = newRolesLocations.( p );
            code.regex = "RPH__" + p;
            replaceAll@StringUtils( code )( code )
          };
          

          response.roles.( newRole ).code[ rci ].code = code
        }
      }
    };

    response.roles.( newRole ).location = newRolesLocations.( newRole ) + "/Activity/basicActivity";
    response.roles.( newRole ).cookie = "basicActivity"
  };

  // checks if the rule has the code for the considered role
  foreach( role : request.ports ) {
    rci = 0; // role code counter
    response.roles.( role ).location = request.ports.( role ).address;
    // WE DO NOT SET THE COOKIE FOR THE ALREADY-PRESENT ROLES BECAUSE THE LEADER HAS THAT COOKIE

    // if it has the code for that role
    if( is_defined( pRoles.( role ) ) ){

      // retrieves the "main" adapted scope and the message-handler
      sf.directory = ruleDir + fileSeparator + role;
      println@Console( "Entering " + sf.directory )();
      sf.regex = ".*\\.ol";
      list@File( sf )( fileList );
      for( fileIndex = 0, fileIndex < #fileList.result, fileIndex++ ){
        file = fileList.result[ fileIndex ];
        println@Console( "Loading file " + role + fileSeparator + file )();
        readFile.filename = sf.directory + fileSeparator + file;
        readFile@File( readFile )( code );

        response.roles.( role ).code[ rci ].key = main_key;

        if( file == "mh.ol" ){
          response.roles.( role ).code[ rci ].mh = code
        } else {
          // if it is the scope, it injects a fresh key related to that scope
          code.replacement = main_key;
          code.regex = "Adapt__KPH__";
          replaceAll@StringUtils( code )( code );
          // replaces the address of the scope leader
          code.replacement = request.client;
          code.regex = "RPH__ADAPTATION_LEADER";
          replaceAll@StringUtils( code )( code );
          // replaces the addresses of the existing roles
          foreach ( p : request.ports ){
            code.replacement = request.ports.( p ).address;
            code.regex = "RPH__" + p;
            replaceAll@StringUtils( code )( code )
          };
          // replaces the addresses of the newRoles
          foreach ( p : newRolesLocations ){
            code.replacement = newRolesLocations.( p );
            code.regex = "RPH__" + p + "!";
            replaceAll@StringUtils( code )( code )
          };
          // replaces the addresses of the newRoles if scope leaders
          foreach ( p : newRolesLocations ){
            code.replacement = newRolesLocations.( p );
            code.regex = "RPH__" + p;
            replaceAll@StringUtils( code )( code )
          };
          // replaces the static UUID keys with the fresh ones
          foreach( staticScopeUUID : subscopesMap ){
            code.replacement = subscopesMap.( staticScopeUUID );
            code.regex = staticScopeUUID;
            replaceAll@StringUtils( code )( code )
          };
          response.roles.( role ).code[ rci ].code = code
        }
      };

      // retrieves the subscopes and the message-handler
      undef( sf.regex );
      sf.dirsOnly = true;
      list@File( sf )( subscopesList );
      undef( sf.dirsOnly );
      sf.regex = ".*\\.ol";
      scopeDirectory = sf.directory;
      for( ssIndex = 0, ssIndex < #subscopesList.result, ssIndex++ ){
        ssFolder = subscopesList.result[ ssIndex ];
        println@Console( "Entering subscope " + ssFolder )();
        sf.directory = scopeDirectory + fileSeparator + ssFolder;
        list@File( sf )( fileList );
        rci++;
        for( fileIndex = 0, fileIndex < #fileList.result, fileIndex++ ){
          file = fileList.result[ fileIndex ];
          println@Console( "Loading file " + sf.directory + fileSeparator + file )();
          readFile.filename = sf.directory + fileSeparator + file;
          readFile@File( readFile )( code );

          response.roles.( role ).code[ rci ].key = subscopesMap.( ssFolder );

          if( file == "mh.ol" ){
            response.roles.( role ).code[ rci ].mh = code
          } else {
            // replaces the static UUID key with the fresh one
            foreach( staticScopeUUID : subscopesMap ){
              println@Console( "replacing key: " + staticScopeUUID + " with " + subscopesMap.( staticScopeUUID ) )();
              code.replacement = subscopesMap.( staticScopeUUID );
              code.regex = staticScopeUUID;
              replaceAll@StringUtils( code )( code )
            };
            // replaces the addresses of the existing roles
            foreach ( p : request.ports ) {
              code.replacement = request.ports.( p ).address;
              code.regex = "RPH__" + p;
              println@Console( "replacing address: " + code.regex + " with " + code.replacement )();
              replaceAll@StringUtils( code )( code )
            };
            // replaces the addresses of the newRoles
            foreach ( p : newRolesLocations ){
              code.replacement = newRolesLocations.( p );
              code.regex = "RPH__" + p + "!";
              replaceAll@StringUtils( code )( code )
            };
            // replaces the addresses of the newRoles if scope leaders
            foreach ( p : newRolesLocations ){
              code.replacement = newRolesLocations.( p );
              code.regex = "RPH__" + p;
              replaceAll@StringUtils( code )( code )
            };

            response.roles.( role ).code[ rci ].code = code
          }
        }
      }
      
    // if there are no projected processes for that role
    } else {
      // loads the empty scope
      response.roles.( role ).code[ rci ].key = main_key;
      code = emptyProcess;
      code.replacement = request.client;
      code.regex = "RPH__ADAPTATION_LEADER";
      replaceAll@StringUtils( code )( code );
      code.replacement = main_key;
      code.regex = "Adapt__KPH__";
      replaceAll@StringUtils( code )( code );
      code.replacement = role;
      code.regex = "RPH__m";
      replaceAll@StringUtils( code )( code );
      foreach ( p : request.ports ) {
        code.replacement = request.ports.( p ).address;
        code.regex = "RPH__" + p;
        replaceAll@StringUtils( code )( code )
      };
      response.roles.( role ).code[ rci ].code = code
    }
  }
}

define checkRules
{
  getVariableNames@AdaptationManager()( envVariableNames );
  rule -> rules[i];
  for( i = 0, i < #rules, i++ ) {
    checkRule;
    if ( send ){
      println@Console( "\tConstraints satisfied" )();
      sendRule;
      i=#rules // rule found, exit the loop
    } else {
      println@Console( "\tConstraint not satisfied, skipping rule" )()
    }
  }
}

main
{
  checkForUpdate( request )( response ) {
    checkRules;
    println@Console( "sending response with " + #response + " roles" )()
  }
}
